/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Clippings extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */


function nsClippingsService() {
  // Public constants
  this.FILETYPE_RDF_XML = 0;

  // Private constants (not declared in interface definition)
  this._SEQNODE_RESOURCE_URI = "http://clippings.mozdev.org/rdf/user-clippings";
  this._PREDNAME_RESOURCE_URI = "http://clippings.mozdev.org/ns/rdf#name";
  this._PREDTEXT_RESOURCE_URI = "http://clippings.mozdev.org/ns/rdf#text";
}

nsClippingsService.prototype = {
  _debug:   false,
  _dataSrc: null,
  _rdfSvc:  Components.classes["@mozilla.org/rdf/rdf-service;1"]
                      .getService(Components.interfaces.nsIRDFService),
  _rdfContainerUtils: Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils),
  _rdfContainer: null,

  get numItems() 
  {
    var rv;
    if (this._rdfContainer) {
      rv = this._rdfContainer.GetCount();
    }
    return rv;
  }
};


nsClippingsService.prototype.getDataSource = function (aDataSrcURL) 
{
  if (this._dataSrc) {
    return this._dataSrc;
  }

  var seqNode = this._rdfSvc.GetResource(this._SEQNODE_RESOURCE_URI);

  try {
    // If Clippings data file doesn't exist, it will be created automatically.
    this._dataSrc = this._rdfSvc.GetDataSourceBlocking(aDataSrcURL);

    // MakeSeq will create a new <RDF:Seq> if it doesn't already exist.
    // If it does, it just initializes it
    this._rdfContainer = this._rdfContainerUtils.MakeSeq(this._dataSrc, seqNode);
    this._rdfContainer = this._rdfContainer.QueryInterface(Components.interfaces.nsIRDFContainer);
  }
  catch (e) {
    this._log(e);
    return null;
  }

  this._log("nsClippingsService.init(): Initialization complete\nDatasource URL: \"" + aDataSrcURL + "\"\n" + this._rdfContainer.GetCount() + " item(s)");
  return this._dataSrc;
};


nsClippingsService.prototype.newClippingFromText = function (aText, aShowDialog, aChromeWnd)
{
  var rv, clipName;
  var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
  var maxNameLen = prefSvc.getIntPref("clippings.entries.max_name_length") || 48;
  var clipText = new String(aText);
  if (clipText && aText) {
    if (clipText.length > maxNameLen) {
      clipName = this._strtrm(clipText).substr(0, maxNameLen) + "...";
    } 
    else {
      clipName = this._strtrm(clipText);
    }

    if (aShowDialog) {
      var args = {name: clipName, text: clipText, userCancel: null};
      aChromeWnd.openDialog("chrome://clippings/content/new.xul", "newclp_dlg",
			    "centerscreen,modal", args);
      if (args.userCancel) {
	return null;
      }
      clipName = args.name;
      clipText = args.text;
    }

    rv = this.createNewClipping(clipName, clipText, aChromeWnd);
  } 
  return rv;
};


nsClippingsService.prototype.createNewClipping = function (aName, aText, aChromeWnd)
{
  var newNode = this._rdfSvc.GetAnonymousResource();
  this._createNewClippingHelper(newNode, aName, aText, null, aChromeWnd);
  return newNode.Value;  // URI of new RDF node
};


nsClippingsService.prototype.createNewClippingWithURI = function (aURI, aName, aText, aPos, aChromeWnd)
{
  var newNode = this._rdfSvc.GetResource(aURI);
  this._createNewClippingHelper(newNode, aName, aText, aPos, aChromeWnd);
  return aURI;
};


nsClippingsService.prototype._createNewClippingHelper = function (aNode, aName, aText, aPos, aChromeWnd)
{
  var predName = this._rdfSvc.GetResource(this._PREDNAME_RESOURCE_URI);
  var predText = this._rdfSvc.GetResource(this._PREDTEXT_RESOURCE_URI);
  var targName = this._rdfSvc.GetLiteral(aName);
  var targText = this._rdfSvc.GetLiteral(aText);

  this._dataSrc.Assert(aNode, predName, targName, true);
  this._dataSrc.Assert(aNode, predText, targText, true);

  if (aPos) {
    this._rdfContainer.InsertElementAt(aNode, aPos, true);
  }
  else {
    this._rdfContainer.AppendElement(aNode);
  }

  if (! aChromeWnd) {
    return;
  }
  // Notify Clippings Manager of the newly-created entry if it is open at the
  // same time as the host app (new entry can be added to Clippings Manager's
  // undo history, etc.)
  // TO DO: Reimplement this using subject-observer pattern
  var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                     .getService(Components.interfaces.nsIWindowMediator);
  if (aChromeWnd.location.href != "chrome://clippings/content/clippings.xul"
      && wm.getMostRecentWindow("ae:clippings-manager")) {
    this.getNewHostAppClippingURI.data = aNode.Value;
  }
};


nsClippingsService.prototype.getNewHostAppClippingURI = function ()
{
  var rv = this.getNewHostAppClippingURI.data;
  this.getNewHostAppClippingURI.data = null;
  return rv;
};

// Static variable for the preceding member function
nsClippingsService.prototype.getNewHostAppClippingURI.data = null;


nsClippingsService.prototype.remove = function (aURI)
{
  var subjectNode = this._rdfSvc.GetResource(aURI);
  var predName = this._rdfSvc.GetResource(this._PREDNAME_RESOURCE_URI);
  var predText = this._rdfSvc.GetResource(this._PREDTEXT_RESOURCE_URI);
  var targName = this._dataSrc.GetTarget(subjectNode, predName, true);
  var targText = this._dataSrc.GetTarget(subjectNode, predText, true);

  try {
    this._dataSrc.Unassert(subjectNode, predName, targName);
    this._dataSrc.Unassert(subjectNode, predText, targText);
    this._rdfContainer.RemoveElement(subjectNode, true);
  } 
  catch (e) {
    this._log(e);
    return false;   
  }

  return true;
};


nsClippingsService.prototype.getName = function (aURI)
{
  var rv;
  var subjName = this._rdfSvc.GetResource(aURI);
  var predName = this._rdfSvc.GetResource(this._PREDNAME_RESOURCE_URI);
  var targName = this._dataSrc.GetTarget(subjName, predName, true);
  if (targName instanceof Components.interfaces.nsIRDFLiteral) {
    rv = targName.Value;
  }
  return rv;
};


nsClippingsService.prototype.getText = function (aURI)
{
  var rv;
  var subjText = this._rdfSvc.GetResource(aURI);
  var predText = this._rdfSvc.GetResource(this._PREDTEXT_RESOURCE_URI);
  var targText = this._dataSrc.GetTarget(subjText, predText, true);
  if (targText instanceof Components.interfaces.nsIRDFLiteral) {
    rv = targText.Value;
  }
  return rv;
};


nsClippingsService.prototype.setName = function (aURI, aName)
{
  var subjName = this._rdfSvc.GetResource(aURI);
  var predName = this._rdfSvc.GetResource(this._PREDNAME_RESOURCE_URI);
  var targOldName = this._dataSrc.GetTarget(subjName, predName, true);
  var targNewName = this._rdfSvc.GetLiteral(aName);
  this._dataSrc.Change(subjName, predName, targOldName, targNewName);

  return true;
};


nsClippingsService.prototype.setText = function (aURI, aText)
{
  var subjText = this._rdfSvc.GetResource(aURI);
  var predText = this._rdfSvc.GetResource(this._PREDTEXT_RESOURCE_URI);
  var targOldText;
  targOldText = this._dataSrc.GetTarget(subjText, predText, true);
  var targNewText = this._rdfSvc.GetLiteral(aText);
  this._dataSrc.Change(subjText, predText, targOldText, targNewText); 

  return true;
};


nsClippingsService.prototype.changePosition = function (aOldPos, aNewPos)
{
  var node = this._rdfContainer.RemoveElementAt(aOldPos, true);
  this._rdfContainer.InsertElementAt(node, aNewPos, true);
  return true;
};


nsClippingsService.prototype.flushDataSrc = function ()
{
  try {
    this._dataSrc.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  }
  catch (e) {
    this._log(e);
    return false;
  }
  this._log("*** Clippings datasource saved to disk ***");
  return true;
};


nsClippingsService.prototype.exportToFile = function (aFileURL, aFileType)
{
  if (aFileType == this.FILETYPE_RDF_XML) {
    var rds = this._dataSrc.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    try {
      rds.FlushTo(aFileURL);
    }
    catch (e) {
      this._log(e);
      return false;
    }
  }
  else {
    return false;
  }

  return true;
};


nsClippingsService.prototype.importFromFile = function (aFileURL, aFileType)
  // Throws an exception if the import failed
  // Returns the number of Clippings entries in the imported file.
{
  if (aFileType != this.FILETYPE_RDF_XML) { 
    throw Components.Exception("File type not supported", Components.result.NS_ERROR_INVALID_ARG);
  }

  var rv;
  var extDataSrc;
  var extRDFContainer = Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
  var extSeqNode = this._rdfSvc.GetResource(this._SEQNODE_RESOURCE_URI);
  try {
    extDataSrc = this._rdfSvc.GetDataSourceBlocking(aFileURL);
    extRDFContainer.Init(extDataSrc, extSeqNode);
  }
  catch (e) {
    throw Components.Exception("Failed to initialize RDF container!");
  }

  if (! this._rdfContainerUtils.IsSeq(extDataSrc, extSeqNode)) {
    throw Components.Exception("Not an RDF Seq container!");
  }

  rv = extRDFContainer.GetCount();
  this._log("External datasource successfully loaded from \"" + aFileURL + "\":\n" + rv + " entries");

  var predName = this._rdfSvc.GetResource(this._PREDNAME_RESOURCE_URI);
  var predText = this._rdfSvc.GetResource(this._PREDTEXT_RESOURCE_URI);
  var targName, targText, newNode;
  var childNodesEnum = extRDFContainer.GetElements();  // Enumerator
  var childNode;
  while (childNodesEnum.hasMoreElements()) {
    childNode = childNodesEnum.getNext();
    if (childNode instanceof Components.interfaces.nsIRDFResource) {
      targName = extDataSrc.GetTarget(childNode, predName, true);
      targText = extDataSrc.GetTarget(childNode, predText, true);

      if (targName instanceof Components.interfaces.nsIRDFLiteral
	  && targText instanceof Components.interfaces.nsIRDFLiteral) {
	var name = targName.Value;
	var text = targText.Value;
      }
      else {
	throw Components.Exception("Name and Text nodes are not valid RDF literals!");
      }

      newNode = this._rdfSvc.GetAnonymousResource();
      this._createNewClippingHelper(newNode, name, text, null, null);
    }
    else {
      throw Components.Exception("Child node of container is not a valid RDF resource!");
    }
  }

  return rv;
};


nsClippingsService.prototype._log = function (aMessage)
{
  if (this._debug) {
    var consoleSvc = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
    consoleSvc.logStringMessage(aMessage);
  }
};


nsClippingsService.prototype._strtrm = function (aString)
{
  const SPACE_CODE = 32;

  if (! aString) {
    return aString;
  }

  if (aString.charCodeAt(0) > SPACE_CODE 
      && aString.charCodeAt(aString.length - 1) > SPACE_CODE) {
	return aString;
      }

  var k = -1;
  while (k < aString.length && aString.charCodeAt(++k) <= SPACE_CODE);
  if (k == aString.length) {
    return "";
  }

  var m = aString.length;
  while (m > 0 && aString.charCodeAt(--m) <= SPACE_CODE);
  return aString.substring(k, m + 1);
};


// nsISupports
nsClippingsService.prototype.QueryInterface = function (aIID)
{
  if (!aIID.equals(Components.interfaces.nsIClippingsService)
      && !aIID.equals(Components.interfaces.nsISupports)) {
	throw Components.results.NS_ERROR_NO_INTERFACE;
      }
  return this;
};



//
// Component registration
//

var Module = {
  cid:       Components.ID("{8ebf6eef-c354-4211-89d0-603b45b76393}"),
  progID:    "clippings@mozdev.org/clippings;1",
  firstTime: true,

  registerSelf: function (aCompMgr, aFileSpec, aLocation, aType) 
  {
  if (this.firstTime) {
      dump("**** First time registration of Clippings component ****\n");
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    dump("**** Registering Clippings component! ****\n");

    aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.cid, "Clippings Component", this.progID, aFileSpec, aLocation, aType);
  },

  getClassObject: function (aCompMgr, aCID, aIID) 
  {
    if (!aCID.equals(this.cid)) {
      throw Components.results.ERROR_NO_INTERFACE;
    }
    if (!aIID.equals(Components.interfaces.nsIFactory)) {
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }
    return this.factory;
  },

  factory: 
  {
    createInstance: function (aOuter, aIID) 
    {
      dump("CI: " + aIID + "\n");
      if (aOuter != null) {
	throw Components.results.NS_ERROR_NO_AGGREGATION;
      }
      return (new nsClippingsService()).QueryInterface(aIID);
    }
  },

  canUnload: function (aCompMgr) 
  {
    dump("**** Unloading: Clippings Component! ****\n");
    return true;
  }
};


function NSGetModule(aCompMgr, aFileSpec) 
{
  return Module; 
}
